//Project: R2 (= Rover number 2)
//Homepage: www.HomoFaciens.de
//Author Norbert Heinz
//Version: 1.0
//Creation date: 09.04.2012
//Copyright: (c) 2012 by Norbert Heinz
//Microcontroller: Atmega8-16PU
//This program is free software you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation version 3 of the License.
//This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
//You should have received a copy of the GNU General Public License along with this program if not, see http://www.gnu.org/licenses/
//
//Program controls 3 servos (steering, camera height, and angle) 1 H-brige (drive motor), one power transistor (light) and one ADC (battery status)
//Program based on PowerSwitch (author Christian Starkjohann) and usb-servo (author Ronald Schaten) - Thanks!
//Program uses V-USB,  published  under GNU GPL v2, GNU GPL v3 or proprietary (CommercialLicense.txt), Copyright: (c) 2005 by OBJECTIVE DEVELOPMENT Software GmbH


#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>

#include "usbdrv/usbdrv.h"
#include "usbdrv/oddebug.h"
#include <util/delay.h>

#define EEPROM_LOCATION (void *)37
#define BATTERYLOW 205

static int powerOn = 0;
static int timerCount02 = 0;
static int steeringAngle = 255;
static int headLight = 0;
static int timeCounter01 = 0;
static int cameraHeight = 0;
static int verticalAngle = 255;
static int controlBits = 0;     //Bit 4 = Forward/reverse drive motor
static int batteryValue = 222;
static int shutdownCounter01 = 0;
static int shutdownCounter02 = 0;
static int verticalServoON = 0;
static int temp01 = 0;

usbMsgLen_t usbFunctionSetup(uchar setupData[8]){
static uchar    replyBuf[2];

  usbMsgPtr = replyBuf;
  usbRequest_t *rq = (void *)setupData;   // cast to structured data for parsing

  switch(rq->bRequest){
    case 0:                                 //ECHO
      replyBuf[0] = rq->wValue.bytes[0];
      replyBuf[1] = rq->wValue.bytes[1];
      return 2;

    case 1:                                 //Set servo steering angle
      if(batteryValue >= BATTERYLOW){
        steeringAngle = setupData[2];
      }
      return 0;

    case 2:                                 //Drive forward for powerOn * 20 ms
      if(batteryValue >= BATTERYLOW){
        PORTB = PORTB & 0b11001111;
        PORTB = PORTB | 0b00010000;
        controlBits &= 0b11101111;
        powerOn = setupData[2];
      }
      return 0;

    case 3:                                 //Drive backward for powerOn * 20 ms
      if(batteryValue >= BATTERYLOW){
        PORTB = PORTB & 0b11001111;
        PORTB = PORTB | 0b00100000;
        controlBits |= 0b00010000;
        powerOn = setupData[2];
      }
      return 0;

    case 4:                                 //Switch headlight
      if(batteryValue >= BATTERYLOW){
        headLight = setupData[2];
      }
      return 0;

    case 5:                                 //Set camera height
      if(batteryValue >= BATTERYLOW){
        cameraHeight = setupData[2];
      }
      return 0;

    case 6:                                 //Set vertical angle of camera
      if(batteryValue >= BATTERYLOW){
        verticalAngle = setupData[2];
        verticalServoON = 2;
      }
      return 0;

    case 10:                                 //Initialize
      if(batteryValue >= BATTERYLOW){
        steeringAngle = 140;
        cameraHeight = 0;
        verticalAngle = 140;
        headLight = 0;
        powerOn = 0;
      }
      return 0;

    case 11:                                 //Turn off servos and lights - sleep mode
      steeringAngle = 0;
      cameraHeight = 0;
      verticalAngle = 0;
      headLight = 0;
      powerOn = 0;
      return 0;

    case 15:                               //Send current battery status
      replyBuf[0] = batteryValue;
      replyBuf[1] = 0;
      return 2;
  }
  return 0;                               // ignore all unknown requests
}

//++++++++++++++++++++ timerInterrupt_0 +++++++++++++++++++++++++++++++
static void timerInterrupt_0(void){

  if(shutdownCounter01<255){
    shutdownCounter01++;
  }
  else{
    shutdownCounter01=0;
    if(batteryValue < BATTERYLOW){
      shutdownCounter02++;
    }  
  }
  
  if(shutdownCounter01 == 50 || shutdownCounter01 == 100 || shutdownCounter01 == 150 || shutdownCounter01 == 200 || shutdownCounter01 == 250 ){
  
    if(verticalServoON > 0){
      verticalServoON--;
    }
    if(batteryValue >= BATTERYLOW){
      if(!(ADCSRA & (1<<ADSC))){//Start new AD conversion if previous one was finished.
        temp01 = (ADCW >> 2);                    // Read and return ADC value as 8 Bit value (!)
        if(temp01 < batteryValue){
          batteryValue--;
        }
        if(temp01 > batteryValue){
          batteryValue++;
        }
        ADMUX = (ADMUX & ~(0x1F)) | (0 & 0x1F);
        ADCSRA |= (1<<ADSC);            //Single conversion
      }
    }

    if(batteryValue < BATTERYLOW){
      PORTC ^= 0b00001000; //Toggle control LED
    }
  }
}

//++++++++++++++++++++ timerInterrupt_1 +++++++++++++++++++++++++++++++
static void timerInterrupt_1(void){

  if(headLight > 0){
    PORTB = PORTB | 0b00001000;
  }

  timeCounter01++;
  if(timeCounter01 >= 5){
    timeCounter01 = 0;
    if(steeringAngle > 0 && steeringAngle < 180){
      PORTB |= 0b00000001;  //Turn servo pins on
    }
    if(verticalAngle > 0 && verticalAngle < 180 && verticalServoON > 0){
      PORTB |= 0b00000010;  //Turn servo pins on
    }
    if((cameraHeight == 1 && (PINC & (1 << 2))) || (cameraHeight == 2 && (PINC & (1 << 1)))){
      PORTB |= 0b00000100;  //Turn servo pins on
    }
    if(powerOn > 0){
      powerOn--;
    }
  }

}

//++++++++++++++++++++ timerInterrupt_2 +++++++++++++++++++++++++++++++
static void timerInterrupt_2(void){
  
  if(powerOn > 0){//Pulse pause modulation to slow down drive motor
    if(timerCount02 == 0){//Turn motor on
      if(!(controlBits & (1 << 4))){//Forward direction
        PORTB &= 0b11011111;
        PORTB |= 0b00010000;
      }
      else{//Reverse direction
        PORTB &= 0b11101111;
        PORTB |= 0b00100000;
      }
    }
    if(timerCount02 == 1){//Turn off motor
      PORTB &= 0b11001111;
    }
  }

  timerCount02++;

  if(timerCount02 == 2){
    timerCount02 = 0;
  }
  
}

void ADC_Init(void) {
 
  uint16_t result;
 
  ADMUX = (1<<REFS1) | (1<<REFS0); // internal reference for ADC

//  ADMUX = (1<<REFS0); // AVCC as refernce for ADC
  
  // Bit of ADFR ("free running") in ADCSRA if zero when truning on, single conversion
  ADCSRA = ((1<<ADPS1) | (1<<ADPS2));     // Frequency didisor 64
  ADCSRA |= (1<<ADEN);                   // Activate ADC
 
  // one "Dummy-Readout" to initiate ADC
 
  ADCSRA |= (1<<ADSC);                  // one conversion
  while (ADCSRA & (1<<ADSC) ) {}        // wait until conversion is complete
  result = ADCW;
}

int main(void){
uchar   i;

  //Port B, Pin 0 = Servo steering angle
  //Port B, Pin 1 = Servo vertical camera angle
  //Port B, Pin 2 = Servo camera height
  //Port B, Pin 3 = Output, Power transistor light
  //Port B, Pin 4 = Output, H-Bridge driving motor
  //Port B, Pin 5 = Output, H-Bridge driving motor

  //Port B, Pin 6 = External oscillator
  //Port B, Pin 7 = External oscillator

  //Port C, Pin 0 = ADC battery voltage
  //Port C, Pin 1 = Switch camera height
  //Port C, Pin 2 = Switch camera height
  //Port C, Pin 3 = Control LED power 

  //Port D, Pin 2 = Input, USB, Int0
  //Port D, Pin 5 = Output, Transistor power supply
  //Port D, Pin 6 = Input, USB D-
  //Port D, Pin 7 = Input, USB D+

  DDRB   = 0b00111111;
  PORTB  = 0b00000000;
  DDRC   = 0b00001000;
  PORTC  = 0b00001110;
  DDRD   = 0b00100001;
  PORTD  = 0b00000000;  //disable pullup resitors at usb-pins


  //Initialize timer 0 (8-bit) - Shutdown counter
  TCCR0 = ( 1 << CS00 ) | ( 1 << CS02 );  //Prescaler 1024, 12 000 000/1024/255 = 48.8Hz = 21.8ms


  //Initialize timer 1 (16-bit) - servos, PWM headlight
  TCCR1A = 0x00;
  TCCR1B = (1<<CS12) | (1<<WGM12); //Configure Timer 1, prescaler 256
  OCR1A = 188;                  //Interrupt after 12,000,000 / 256 / 188 = 249Hz (after 4.011ms)
  OCR1B = 188;                  //Interrupt after 12,000,000 / 256 / 188 = 249Hz (after 4.011ms)

  //Initialize timer 2 (8-bit) - IR communication
  OCR2 = 15;       //Compare value
  TCCR2 = ( 1 << WGM21 ) | ( 1 << CS22 );  //Switch to CTC mode, prescaler 64, 12 000 000/64/15 = 12500Hz = 0.08ms


  PORTB = PORTB | 0b00001000; //Turn on Light for a short span to indicate that microcontroller is initializing

  usbDeviceDisconnect();  /* enforce re-enumeration, do this while interrupts are disabled! */
  i = 0;
  while(--i){             /* fake USB disconnect for > 500 ms */
    wdt_reset();
    _delay_ms(2);
  }
  usbDeviceConnect();
  usbInit();
  sei();

  ADC_Init();

  PORTD |= 0b00100000; //Turn ON PowerSupply !!!
  PORTC |= 0b00001000; //Turn ON control LED

  PORTB = PORTB & 0b11110111; //Turn off light after initialisazion


  while(1 == 1){    /* main event loop */
    usbPoll();
    
    if(TCNT1 > steeringAngle){
      PORTB &= 0b11111110;  //Turn servo pin off
    }
    if(TCNT1 > verticalAngle){
      PORTB &= 0b11111101;  //Turn servo pin off
    }
    if ((cameraHeight == 1 && TCNT1 > 50) || (cameraHeight == 2 && TCNT1 > 150)){
      PORTB &= 0b11111011;  //Turn servo pin off
    }
    if(TCNT1 > headLight){
      PORTB &= 0b11110111;  //headlight off
    }
    
    if(batteryValue < BATTERYLOW){
      steeringAngle = 0;
      cameraHeight = 0;
      verticalAngle = 0;
      headLight = 0;
      powerOn = 0;      
    }

    if(shutdownCounter02 > 90){
      PORTD &= 0b11011111; //Turn OFF Power Supply !!!
    }
    
    
    if (TIFR & (1 << TOV0)) {//'Manual' timer interrupt. Normal timer interrupt troubles USB-communication
      TIFR |= 1 << TOV0;  /* clear pending flag */
      TCNT0 = 0;
      timerInterrupt_0();
    }

    if (TIFR & (1 << OCF1A) && !(controlBits & (1 << 0))) {//'Manual' timer interrupt. Normal timer interrupt troubles USB-communication
      TIFR |= 1 << OCF1A;  /* clear pending flag */
      TCNT1 = 0;
      timerInterrupt_1();
    }

    if (TIFR & (1 << OCF2) ) {//'Manual' timer interrupt. Normal timer interrupt troubles USB-communication
      TIFR |= 1 << OCF2;  /* clear pending flag */
      TCNT2 = 0;
      timerInterrupt_2();
    }

    if(powerOn == 0){//Turn off drive motor
      PORTB = PORTB & 0b11001111;
    }

  }//Main loop
  return 0;
}

